// DCConfigAPI.h: API interface for DC configuration reader/write classes

/*
// CodeAnalyst for Open Source
// Copyright 2006 Advanced Micro Devices, Inc.
// You may redistribute this program and/or modify this program under the terms
// of the GNU General Public License as published by the Free Software 
// Foundation; either version 2 of the License, or (at your option) any later 
// version.
//
// This program is distributed WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  See the 
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA 02111-1307 USA.
*/
#ifndef _DCCONFIGAPI_H_
#define _DCCONFIGAPI_H_

// Include Qt-related definitions
#include <qstring.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qxml.h>
#include <qvaluelist.h>

#include <list>
using namespace std ;

//
// Linux
// DCCONFIGUREFILE_EXPORTS should be defined to be empty
//
#define DCCONFIGUREFILE_API

//
// Number of event counters per group. This is a platform-dependent
// constant which may somebody be defined in a central location!
//
static const int NumberOfCounters = 4 ;

typedef struct
{
	int numberOfValidEvents ;                      // Valid events in the group
	unsigned int eventSelect[NumberOfCounters] ;   // Event select
	unsigned int eventUnitMask[NumberOfCounters] ; // Event unit mask
	UINT64 eventCount[NumberOfCounters] ;          // Event count aka "sampling period"
	bool eventOS[NumberOfCounters] ;               // Event OS sampling control
	bool eventUser[NumberOfCounters] ;             // Event User sampling control
	bool edgeDetect[NumberOfCounters] ;            // Event edge detect control
	bool eventHost[NumberOfCounters] ;             // Event Host sampling control
	bool eventGuest[NumberOfCounters] ;            // Event Guest sampling control
} EventGroup, *pEventGroup ;

// Do not use the following type in DLL client code
typedef list<EventGroup> EventGroupList ;          // Causes warning C4251

typedef enum
{
	DCConfigInvalid,           // No or invalid DC configuration
	DCConfigTBP,               // Time-based profiling (TBP)
	DCConfigEBP,               // Event-based profiling (EBP)
	DCConfigSim,               // Pipeline simulation (Sim)
	DCConfigIBS                // Instruction-based sampling (IBS)
} DCConfigType ;

typedef struct _SimConfig
{
	bool warmupCaches ;        // Warm up caches before simulation
	bool runSimAfter ;         // Run simulation after tracing
	bool saveTrace ;           // Save generated trace file
	QString procCore ;         // Processor to use for simulation
	QString traceFileName ;    // Name of trace file
	unsigned int multiplier ;  // Base clock multiplier
	unsigned int maxToTrace ;  // Max number of instructions to trace
} SimConfig, *pSimConfig ;

typedef struct _IBSConfig
{
	UINT64 fetchMaxCount ;     // Maximum value of periodic fetch counter
	UINT64 opMaxCount ;        // Maximum value of periodic op counter
	bool fetchSampling ;       // Enable IBS fetch sampling
	bool opSampling ;          // Enable IBS op sampling
	bool opCycleCount;         //Whether the op sampling is by cycle or dispatch
	bool dcMissInfoEnabled;
	QValueList<unsigned int> fetchList;
	QValueList<unsigned int> opList;
} IBSConfig, *pIBSConfig ;

////////////////////////////////////////////////////////////////////////////////////
// CDCConfig class declaration
////////////////////////////////////////////////////////////////////////////////////

class DCCONFIGUREFILE_API CDCConfig : public QXmlDefaultHandler
{
public:
	///////////////////////////////////////////////////////////////////////////////
	// Constructors, destructors, assignment
	///////////////////////////////////////////////////////////////////////////////
	CDCConfig() ;
	CDCConfig(const CDCConfig& original) ;
	~CDCConfig() ;
	const CDCConfig& operator = (const CDCConfig& rhs) ;

	///////////////////////////////////////////////////////////////////////////////
	// Client functions for all DC configuration types
	///////////////////////////////////////////////////////////////////////////////
	// Return the configuration type (TBP, EBP, ...)
	DCConfigType GetConfigType() ;
	// Return the configuration name in configName
	void GetConfigName(QString& configName) ;
	// Return the configuration tool type in toolTip
	void GetToolTip(QString& toolTip) ;
	// Return the configuration description in description
	void GetDescription(QString& description) ;

	// Set the configuration type
	void SetConfigType(DCConfigType configType) ;
	// Set the configuration name
	void SetConfigName(const QString configName) ;
	// Set the configuration tool tip
	void SetToolTip(const QString toolTip) ;
	// Set the configuration description
	void SetDescription(const QString description) ;

	// Read a DC configuration in XML from the specified file
	bool ReadConfigFile(wchar_t* configFileName) ;
	// Write the current configuration in XML to the specified file
	bool WriteConfigFile(wchar_t* configFileName) ;

	///////////////////////////////////////////////////////////////////////////////
	// Client functions for TBP DC configurations
	///////////////////////////////////////////////////////////////////////////////
	// Set the TBP timer interval (sample period)
	void SetTimerInterval(float interval) ;
	// Get the TBP timer interval
	float GetTimerInterval() ;

	///////////////////////////////////////////////////////////////////////////////
	// Client functions for EBP DC configurations
	///////////////////////////////////////////////////////////////////////////////
	// Set event groups and multiplex period for the configuration
	void SetEventInfo(EventGroup* pEventGroup, int numberOfGroups,
		UINT64 multiplexPeriod = 0) ;
	// Return the number of event groups in the configuration
	int GetNumberOfGroups() ;
	// Get the event groups; Copy data to caller-supplied array of event groups
	void GetEventInfo(EventGroup* pEventGroup, int numberOfGroups) ;
	// Get the multiplexing period; Multiplexing period is zero when disabled
	UINT64 GetMultiplexPeriod() ;

	///////////////////////////////////////////////////////////////////////////////
	// Client functions for SIM DC configurations
	///////////////////////////////////////////////////////////////////////////////
	// Set simulation configuration info using the specified struct
	void SetSimInfo(SimConfig* pSim) ;
	// Get simulation configuration info; Copy info to caller-supplied struct
	void GetSimInfo(SimConfig* pSim) ;

	///////////////////////////////////////////////////////////////////////////////
	// Client functions for IBS DC configurations
	///////////////////////////////////////////////////////////////////////////////
	// Set IBS configuration info using the specified struct
	void SetIBSInfo(IBSConfig* pIBS) ;
	// Get IBS configuration info; Copy info to caller-supplied struct
	void GetIBSInfo(IBSConfig* pIBS) ;

	///////////////////////////////////////////////////////////////////////////////
	// Callback functions which override QXmlDefaultHandler
	///////////////////////////////////////////////////////////////////////////////
	bool startDocument() ;
	bool endDocument() ;
	bool startElement(
		const QString& namespaceURI, const QString& localName,
		const QString& qName, const QXmlAttributes& atts) ;
	bool endElement(
		const QString& namespaceURI, const QString& localName,
		const QString& qName) ;
	bool characters(const QString& ch) ;

private:
	// Common configuration attributes
	DCConfigType     m_configType ;        // Kind of configuration (TBP, EBP, ...)
	QString          m_configName ;        // Configuration name
	QString          m_toolTip ;           // Configuration's tool tip
	QString          m_description ;       // Description of the configuration
	// Time-based profiling configuration attributes
	float            m_interval ;          // Timer interval (sample period)
	// Event-based profiling configuration attributes
	UINT64           m_multiplexPeriod ;   // Multiplexing period (0: no multiplexing)
	int              m_numberOfGroups ;    // Number of event groups
	EventGroup      *m_eventGroups ;       // Points to array of event groups
	// Simulation configuration attributes
	SimConfig        m_simConfig ;         // Holds Sim-specific configuration info
	// IBS configuration attributes
	IBSConfig        m_IBSConfig ;         // Holds IBS-specific configuration info

	// Private data members for XML handling
	EventGroupList   m_eventGroupList ;    // Holds event groups as they come in
	bool             m_toolTipIsOpen ;     // Tooltip is currently open when true
	bool             m_descriptionIsOpen ; // Description is currently open
	bool             m_groupIsOpen ;       // Event group is currently open
	int              m_eventIndex ;        // Array index to store event information
	bool             m_xmlSemanticError ;  // Is true on XML semantic error

	///////////////////////////////////////////////////////////////////////////////
	// Private functions to assist configuration writing
	///////////////////////////////////////////////////////////////////////////////
	// Write tool tip element to the XML stream
	void WriteTooltip(QTextStream& xmlStream, QString& text) ;
	// Write description element to the XML stream
	void WriteDescription(QTextStream& xmlStream, QString& text) ;
	// Write string-valued attribute to the XML stream
	void WriteStringAttr(QTextStream& xmlStream, const char* attr, QString& value) ;
	// Write integer-valued attribute (decimal format) to the XML stream
	void WriteDecimalAttr(QTextStream& xmlStream, const char* attr,
		unsigned long long int value) ;
	// Write integer-valued attribute (hexadecimal format) to the XML stream
	void WriteHexAttr(QTextStream& xmlStream, const char* attr,
		unsigned long long int value) ;
	// Write float-valued attribute (decimal format) to the XML stream
	void WriteFloatAttr(QTextStream& xmlStream, const char* attr, float value) ;
	// Write Boolean-valued attribute to the XML stream
	void WriteBoolAttr(QTextStream& xmlStream, const char* attr, bool value) ;
	// Write an event group element to the XML stream
	void WriteEventGroup(QTextStream& xmlStream, EventGroup* pEG) ;
	// Write TBP configuration to the XML stream
	void WriteTBP(QTextStream& xmlStream) ;
	// Write EBP configuration to the XML stream
	void WriteEBP(QTextStream& xmlStream) ;
	// Write Sim configuration to the XML stream
	void WriteSim(QTextStream& xmlStream) ;
	// Write IBS configuration to the XML stream
	void WriteIBS(QTextStream& xmlStream) ;
} ;

#endif
