/*
// CodeAnalyst for Open Source
// Copyright 2006 Advanced Micro Devices, Inc.
// You may redistribute this program and/or modify this program under the terms
// of the GNU General Public License as published by the Free Software 
// Foundation; either version 2 of the License, or (at your option) any later 
// version.
//
// This program is distributed WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  See the 
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA 02111-1307 USA.
*/


#include <qradiobutton.h>
#include <qfiledialog.h>
#include <qcheckbox.h>
#include <stdlib.h>

#include "ImportWizard.h"
#include "application.h"
#include "eventsfile.h"
#include "time.h"
#include "atuneoptions.h"
#include "atunefile.h"
#include "ProcessFilterDlg.h"

#define PACKED_DIR_NAME "capacked"
#define CA_DEFAULT_IMPORT_PRD_PATH "/var/lib/oprofile/anon_samples.prd"

#define SHOW_ERR_AND_RETURN  if (NO_ERR != ret) { \
	show_error_message(ret); \
	return; }

ImportWizard::ImportWizard(QWidget* parent, const char* name, bool modal, WFlags fl)
: IImportWizard (parent, name, modal, fl)
{
	m_cpuinfo_path = "/proc/cpuinfo";
	m_dispatched_ops_path = "/dev/oprofile/ibs_op/dispatched_ops";  // Should be empty if doesnot exist.
	set_up_pages();
	load_settings();
}


ImportWizard::~ImportWizard()
{

}

void ImportWizard::set_up_pages()
{
	helpButton()->setHidden(true);

	setNextEnabled(page(LOCAL_PAGE_0), false);
	setNextEnabled(page(REMOTE_PAGE_0), false);
	setNextEnabled(page(EBP_PAGE_0), false);
	setNextEnabled(page(XML_PAGE_0), false);

	setFinishEnabled(page(LOCAL_PAGE_0), true);
	setFinishEnabled(page(REMOTE_PAGE_0), true);
	setFinishEnabled(page(EBP_PAGE_0), true);
	setFinishEnabled(page(XML_PAGE_0), true);

	// We no longer allow user to specify option for import
	m_ck_exclude_dep->hide();
	m_ck_merge_lib->hide();


        m_ebp_path->setText(QDir::home().path() + "/AMD/CodeAnalyst");
        m_xml_path->setText(QDir::home().path() + "/AMD/CodeAnalyst");

}

bool ImportWizard::is_remote_import()
{
	return m_rd_remote->isChecked();	
}

bool ImportWizard::is_ebp_import()
{
	return m_rd_ebp->isChecked();	
}

bool ImportWizard::is_xml_import()
{
	return m_rd_xml->isChecked();	
}

void ImportWizard::load_settings()
{
	unsigned int import_type = LOCAL_PROFILE;
	CATuneOptions ao;

	ao.getImportType(&import_type);
	if (LOCAL_PROFILE == import_type)
		m_rd_local->setChecked(true);
	else if (REMOTE_PROFILE == import_type)
		m_rd_remote->setChecked(true);
	else if (EBP_PROFILE == import_type)
		m_rd_ebp->setChecked(true);
	else if (XML_PROFILE == import_type)
		m_rd_xml->setChecked(true);

	QString op_dir;
	ao.getImportOpdir(op_dir);
	m_op_dir_path->setText(op_dir);

	QString tar;
	ao.getImportTAR(tar);
	m_tar_path->setText(tar);

	unsigned int value;
	ao.getImportExcludeDep(&value);
	m_ck_exclude_dep->setChecked(value);

	ao.getImportMergeByLib(&value);
	m_ck_merge_lib->setChecked(value);
}


void ImportWizard::next()
{
	CATuneOptions ao;

	switch(indexOf(currentPage())) {
	case WIZARD_PAGE_0: 
		if (m_rd_remote->isChecked()) {
			setFinishEnabled(page(REMOTE_PAGE_0), true);
			showPage(page(REMOTE_PAGE_0));
		} else if (m_rd_local->isChecked()) {
			m_cpuinfo_path = "/proc/cpuinfo";
			setFinishEnabled(page(LOCAL_PAGE_0), true);
			showPage(page(LOCAL_PAGE_0));
		} else if (m_rd_ebp->isChecked()) {
			setFinishEnabled(page(EBP_PAGE_0), true);
			showPage(page(EBP_PAGE_0));
		} else if (m_rd_xml->isChecked()) {
			setFinishEnabled(page(XML_PAGE_0), true);
			showPage(page(XML_PAGE_0));
		}
	break;
	}
}


void ImportWizard::back()
{
	switch(indexOf(currentPage())) {
		case LOCAL_PAGE_0:
		case REMOTE_PAGE_0:
		case EBP_PAGE_0:
		case XML_PAGE_0:
			setApplyProcessFilter(false);
			setFinishEnabled(currentPage(), false);
			showPage(page(WIZARD_PAGE_0));
			break;
	}
}

// JNC and Oprofile sample directory must exist
unsigned int  ImportWizard::validate_path()
{
	// Sanity check for empty path.
	if(m_op_sample_path.isEmpty())
		return OP_SAMPLE_DIR_NOT_EXIST;

	QDir dir(m_op_sample_path);

	// Sanity check for not exist path.
	if (!dir.exists())
		return OP_SAMPLE_DIR_NOT_EXIST;

	// Sanity check to see if this directory really contains oprofile samples
	// by checking for directory {root} or {kern}
	if (!dir.exists("{root}") && !dir.exists("{kern}"))
	{
		return OP_INVALID_SAMPLE_DIR; 
	}

	return NO_ERR; 
}

void ImportWizard::show_error_message(unsigned int code)
{
	QString message;
	QString message2;
	QWidget * item = NULL;

	switch(code) {
		case OP_SAMPLE_DIR_NOT_EXIST:
			message = QString("Path \"") + m_op_sample_path + "\" does not exist.";
			item = m_op_dir_path;
			break;
		case OP_SAMPLE_NOT_FOUND:
			message = "Package contains no OProfile sample file."; 
			item = m_tar_path;
			break;
		case OP_INVALID_SAMPLE_DIR:
			message = QString("Invalid Importing Path.\n") + 
				"Please specify full-path to Oprofile sample directory.\n" +
				"It should contains subdirectory \"{kern}\" and/or \"{root}\"." ;
			item = m_op_dir_path;
			break;
		case EBP_EMPTY_PATH:
			message = QString("Invalid Importing Path.\n") + 
				"Please specify full-path to EBP/TBP file.\n" +
				"For Java profile, please also specify JIT directory.\n" +
				"Otherwise, please uncheck Java Profile button.\n";
			item = m_ebp_path;
			break;
		case XML_INVALID_PATH:
			message = QString("Invalid XML Path.\n") + 
				"Please specify full-path to XML file.\n";
			item = m_xml_path;
			break;
		case XML_CPUINFO_INVALID_PATH:
			message = QString("Invalid Cpuinfo Path.\n") + 
				"Please specify full-path to cpuinfo file.\n" +
				"(i.e. /proc/cpuinfo)\n";
			item = m_xml_path;
			break;
	}

	switch(indexOf(currentPage())) {
		case LOCAL_PAGE_0:
			message2 = "Please select a different directory.";
			break;
		case REMOTE_PAGE_0:
			message2 = "The packaged file is missing critical directory.\n" 
				"Please choose a different .tar.gz file or cancel.";
			break;
	}


	message = message + " \n" + message2;
	QMessageBox::critical (this, "Error", message);
	if (item != NULL)
		item->setFocus(); 
}

void ImportWizard::accept()
{
	CATuneOptions ao;
	unsigned int value;

	// Check Process Filter
	if(isApplyProcessFilter())
	if( m_filterList.size() == 0 
	||  (m_filterList.size() == 1 && m_filterList[0].isEmpty() )) 
	{
		unsigned int answer = 	
		 QMessageBox::warning(this,QString("CodeAnalyst Warning"),
                        QString("Apply Process Filter is selected, but Process Filter List is empty.\n")
                        	+ "CodeAnalyst will disable process filtering.\n"
                                + "Are you sure you want to continue?\n",
                        QMessageBox::Yes, QMessageBox::No);
		if(answer != QMessageBox::Yes )
                {
			// Set focus
			if(m_pApplyFilter->isChecked())
				m_pApplyFilter->setFocus();
			
			return;	
		}else{
			setApplyProcessFilter(false);		
		}
	}

	switch(indexOf(currentPage())) {
		case LOCAL_PAGE_0:
			m_op_sample_path = m_op_dir_path->text();
			int ret; 
			if (NO_ERR != (ret = validate_path())) {
				show_error_message(ret);
				return;
			}
			ao.setImportOpdir(m_op_sample_path);

			if (m_ck_merge_lib->isChecked()) {
				value = MERGE_LIB;
				m_op_merge_lib = true;
			}
			else {
				value = NO_MERGE_LIB;
				m_op_merge_lib = false;
			}

			ao.setImportMergeByLib(value);
			ao.getImportMergeByLib(&value);

			if (m_ck_exclude_dep->isChecked())
				value = EXCLUDE_DEP;
			else
				value = NO_EXCLUDE_DEP;

			ao.setImportExcludeDep(value);
			ao.getImportExcludeDep(&value);
			break;
		case REMOTE_PAGE_0:
			m_packed_path = m_tar_path->text();
			ao.setImportTAR(m_packed_path);

			ret = import_packed_file();
			SHOW_ERR_AND_RETURN  

				ret = validate_path();
			SHOW_ERR_AND_RETURN  

			break;
		case EBP_PAGE_0:
			// Handling the EBP/TBP file
			if ( m_ebp_path->text().isEmpty()
			|| !QFile::exists(m_ebp_path->text()) 
			|| ( m_enable_java_import->isChecked() && m_jit_dir_path->text().isEmpty() ))
			{
				show_error_message(EBP_EMPTY_PATH);
				return;	
			}

			// Check if empty

			break;
		case XML_PAGE_0:
			// Handling the XML file
			if (m_xml_path->text().isEmpty() 
			|| !QFile::exists(m_xml_path->text())
			|| !QFileInfo(m_xml_path->text()).isFile()) 
			{
				show_error_message(XML_INVALID_PATH);
				return;	
			}

			if (m_xml_cpuinfo_path->text().isEmpty() 
			|| !QFile::exists(m_xml_cpuinfo_path->text())
			|| !QFileInfo(m_xml_cpuinfo_path->text()).isFile()) 
			{
				show_error_message(XML_CPUINFO_INVALID_PATH);
				return;	
			}
			break;
		default:
			break;
	}

	if (m_rd_local->isChecked())
		value = LOCAL_PROFILE;
	else if (m_rd_remote->isChecked())
		value = REMOTE_PROFILE;
	else if (m_rd_ebp->isChecked())
		value = EBP_PROFILE;
	else if (m_rd_xml->isChecked())
		value = XML_PROFILE;
    
	ao.setImportType(value);

	QDialog::accept();
}

void ImportWizard::browse_packed()
{
	QString packed_path = QFileDialog::getOpenFileName (m_tar_path->text(), "*.tar.gz",
		this, NULL,
                                               "Choose a capackage file");

	if (!packed_path.isEmpty()) 
		m_tar_path->setText(packed_path);
}


void ImportWizard::browse_opsample_dir()
{
	QString path;

	path = QFileDialog::getExistingDirectory(m_op_dir_path->text(), this,
		NULL,
		"Choose a directory containing Oprofile Session");

	if (!path.isEmpty()) 
		m_op_dir_path->setText(path);
}

void ImportWizard::browse_ebp_dir()
{
    QString ebp_path;

    ebp_path = QFileDialog::getOpenFileName(m_ebp_path->text(), "CodeAnalyst Event File (*.tbp *.ebp)",
                this, NULL,
                "Choose an EBP/TBP file.");
    
    if (!ebp_path.isEmpty()) 
        m_ebp_path->setText(ebp_path);
}

void ImportWizard::browse_xml_dir()
{
    QString xml_path;

    xml_path = QFileDialog::getOpenFileName(m_xml_path->text(), "CodeAnalyst Event File (*.xml)",
                this, NULL,
                "Choose an Oprofile XML file.");
    
    if (!xml_path.isEmpty()) 
        m_xml_path->setText(xml_path);
}

void ImportWizard::browse_xml_cpuinfo()
{
    QString path;

    path = QFileDialog::getOpenFileName(m_xml_cpuinfo_path->text(), "cpuinfo file",
                this, NULL,
                "Choose cpuinfo file.");
    
    if (!path.isEmpty()) 
        m_xml_cpuinfo_path->setText(path);
}

void ImportWizard::browse_jit_dir()
{
    QString path;

    path = QFileDialog::getExistingDirectory(m_jit_dir_path->text(), this,
                NULL,
                "Choose a directory containing JIT files");
    
    if (!path.isEmpty()) 
        m_jit_dir_path->setText(path);
}

void ImportWizard::onToggleJavaImport()
{
	// Enable Disable the button group
	if(m_enable_java_import->isChecked())
	{
		m_jit_dir_path->setEnabled(true);
		m_browse_jit_dir->setEnabled(true);
	}else{
		m_jit_dir_path->setEnabled(false);
		m_browse_jit_dir->setEnabled(false);
	}
}

void ImportWizard::default_clicked()
{
	m_op_dir_path->setText(OP_SAMPLES_CURRENT_DIR);
}

unsigned int ImportWizard::import_packed_file()
{
	untar_packed_file();
	return rename_files();
}

void ImportWizard::mkdir_p(QString & absPath)
{
	QStringList list =  QStringList::split("/", absPath);
	QStringList::iterator it = list.begin();

	QString dir_path = "/";
	for (; it != list.end(); ++it) {
		dir_path = dir_path + (*it) + "/";
		QDir dir(dir_path);

		if (!dir.exists())  {
			if (!dir.mkdir(dir_path))
				qDebug("mkdir failed");
		}

	}
}

void ImportWizard::copy_files(QFileInfo & info, QString & dest_dir)
{
	QUrlOperator urlop;

	// Copy any non-directory file to the new path
	QDir dir = QDir(info.filePath());
	dir.setFilter(QDir::Files);
	QStringList file_list = dir.entryList(); 
	QStringList::Iterator itFile = file_list.begin();
	for (; itFile != file_list.end(); ++itFile) {
		QFileInfo info(dir, (*itFile));

		QString command = "cp \"" + info.absFilePath() + "\" \"" + dest_dir + "\"";

		system(command.data());
	}
}


void ImportWizard::recurse_dirs(QString & parent_dir_str, 
								QString & dest_dir_str, QString & subs_dir)
{
	QDir current_dir = QDir(parent_dir_str);   

	current_dir.setFilter(QDir::Dirs);
	QStringList old_dir_list = current_dir.entryList();
	QStringList::Iterator it = old_dir_list.begin();

	for (; it != old_dir_list.end(); ++it) {
		if (*it == "." || *it == "..")
			continue;

		QFileInfo info(current_dir, *it);
		QString path = info.filePath(); 

		QString temp = dest_dir_str + "/" + *it;

		if ((*it).contains("{dep}")) {
			prev_dep = true;
		}
		else if ((*it).contains("{root}") && prev_dep == true) {
			temp += subs_dir;
		} else
			prev_dep =false;

		//        mkdir_p(dest_dir_str);
		mkdir_p(temp);
		copy_files(info, temp);
		recurse_dirs(path, temp, subs_dir);
	}
}


unsigned int ImportWizard::rename_files()
{
	//todo capacked is hard coded right now.
	QString top_dir = m_temp_dir;
	QString op_current_dir_str = m_temp_dir + PACKED_DIR_NAME + "/current";
	QDir op_current_dir(op_current_dir_str );

	// Move the sample file to correct directory structure
	// For example: sample files in  
	// /tmp/CA_982/capacked/current/{root}/bin/basename 
	// needs to be moved to 
	// /tmp/CA_982/capacked/current/{root}/tmp/CA_982/capacked/binary/bin/basename/

	// Since we will be recursing capacked/current directory tree tree
	// the new structure is crecated at at capacked/new_current
	QString dest_topdir_str = top_dir + "capacked/samples/current/{root}" 
		+ m_temp_dir + "capacked/binary";

	m_op_sample_path = top_dir + "capacked/samples/current/";

	m_cpuinfo_path = top_dir + "capacked/cpuinfo";

	m_dispatched_ops_path = top_dir + "capacked/cainfo/dispatched_ops";
	
	QString subs_dir = m_temp_dir + "capacked/binary/{root}/";

	// Loop through the original directory structure and create 
	// correspondent directory under the new directory tree structure
	recurse_dirs(op_current_dir_str, dest_topdir_str, subs_dir);

	return NO_ERR;
}

void ImportWizard::untar_packed_file()
{
	QDir dir;
	QString message = m_packed_path + "does not exist";

	if (!QFile::exists(m_packed_path)) {
		return;    
	}

	srand(time(0));
	m_temp_dir.sprintf("/tmp/CA%ld/", random());
	dir.mkdir(m_temp_dir);

	// todo hook the output to the dialog
	// todo use a randomly generated directory
	QString command = "tar zxf " + m_packed_path + " -C " + m_temp_dir;
	system(command.data());
}

void ImportWizard::find_prd_files(QString & dir_name)
{
	QDir dir(dir_name);

	QStringList file_list = dir.entryList(); 
	QStringList::iterator it = file_list.begin();
	for (; it != file_list.end(); ++it) {
		QFileInfo info(dir, *it);
		QString path = info.filePath(); 

		if (*it == "." || *it == "..")
			continue;

		if (path.contains(".prd")) {
			m_packed_prd_path = path;
			return;
		}

		find_prd_files(path);
	}
}


QString ImportWizard::get_packed_prd_path()
{
	return m_packed_prd_path; 
}

QString ImportWizard::get_ebp_path()
{
   return m_ebp_path->text(); 
}

QString ImportWizard::get_xml_path()
{
   return m_xml_path->text(); 
}

QString ImportWizard::get_xml_cpuinfo_path()
{
   return m_xml_cpuinfo_path->text(); 
}

QString ImportWizard::get_jit_dir_path()
{
   return m_jit_dir_path->text(); 
}

QString ImportWizard::get_op_sample_dir()
{
	return m_op_sample_path;
}


QString ImportWizard::get_cpuinfo_path()
{
	return m_cpuinfo_path; 
}

QString ImportWizard::get_cainfo_dispatched_ops_path()
{
	return m_dispatched_ops_path; 
}


bool ImportWizard::get_op_merge_lib()
{
	return m_op_merge_lib;
}


bool ImportWizard::get_exclude_dep()
{
	return m_op_exclude_dep;
}

void ImportWizard::onToggleApplyFilter(bool t)
{
	setApplyProcessFilter(t);
	
	// We pop the Process Filter Dialog automatically.
	if(t)
	{
		onClickApplyFilterAdvanceBtn();
	}

}

void ImportWizard::onClickApplyFilterAdvanceBtn()
{
	ProcessFilterDlg *pfd = new ProcessFilterDlg(this, "");
	
	pfd->init(m_filterList);
	if( QDialog::Accepted == pfd->exec())
	{
		m_filterList.clear();
		m_filterList = pfd->getStringList();
	}

	if(pfd)
	{
		delete pfd;
		pfd = NULL;
	}
}

bool ImportWizard::isApplyProcessFilter()
{
	return m_pApplyFilter->isChecked();
}

QStringList ImportWizard::getProcessFilter()
{
	return m_filterList;
}

void ImportWizard::setProcessFilter(QStringList list)
{
	m_filterList = list;
}

void ImportWizard::setApplyProcessFilter(bool t)
{
	m_pApplyFilter->setChecked(t);
	m_pApplyFilterAdvanceBtn->setEnabled(t);
}
